##############################################################################
#
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
#
# Copyright (c) Digital Creations.  All rights reserved.
#
# This license has been certified as Open Source(tm).
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
#
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
#
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
#
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
#
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
#
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
#
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
#
##############################################################################
__doc__='''Define Zope\'s default security policy


$Id: DebugSecurityPolicy.py,v 1.7 2001/02/12 14:49:30 shane Exp $'''
__version__='$Revision: 1.7 $'[11:-2]

from AccessControl import SimpleObjectPolicies
try:
    _noroles = SimpleObjectPolicies._noroles
except:
    _noroles = []

from AccessControl.PermissionRole import \
     _what_not_even_god_should_do, rolesForPermissionOn
from string import join
from Acquisition import aq_inner, aq_parent, aq_base

from utils import getPathOf

class DebugSecurityPolicy:

    def validate(self, accessed, container, name, value, context,
                 roles=_noroles, None=None, type=type, IntType=type(0),
                 DictType=type({}), getattr=getattr, _noroles=_noroles,
                 StringType=type(''),
                 Containers=SimpleObjectPolicies.Containers,
                 valid_aq_=('aq_parent','aq_explicit')):

        ############################################################
        # Provide special rules for the acquisition attributes
        if type(name) is StringType:
            if name[:3]=='aq_' and name not in valid_aq_:
                info = 'aq_* names not allowed.'
                logProtection(info, 0)
                return 0

        containerbase=aq_base(container)
        accessedbase=aq_base(accessed)

        ############################################################
        # Try to get roles
        if roles is _noroles:
            roles=getattr(value, '__roles__', _noroles)

        if roles is _noroles:

            ############################################################
            # We have an object without roles. Presumably, it's
            # some simple object, like a string or a list.
            if container is None:
                info = 'Unspecified container of simple object.'
                logProtection(info, 0)
                return 0 # Bail if no container

            roles=getattr(container, '__roles__', _noroles)
            if roles is _noroles:
                aq=getattr(container, 'aq_acquire', None)
                if aq is None:
                    roles=_noroles
                    if containerbase is not accessedbase:
                        info = 'Container is not an acquisition wrapper.'
                        logProtection(info, 0)
                        return 0
                else:
                    # Try to acquire roles
                    try: roles=aq('__roles__')
                    except AttributeError:
                        roles=_noroles
                        if containerbase is not accessedbase:
                            info = 'Unable to find __roles__ in container.'
                            logProtection(info, 0)
                            return 0

            # We need to make sure that we are allowed to
            # get unprotected attributes from the container. We are
            # allowed for certain simple containers and if the
            # container says we can. Simple containers
            # may also impose name restrictions.
            p=Containers(type(container), None)
            if p is None:
                p=getattr(container,
                          '__allow_access_to_unprotected_subobjects__', None)

            if p is not None:
                tp=type(p)
                if tp is not IntType:
                    if tp is DictType:
                        p=p.get(name, None)
                    else:
                        p=p(name, value)

            if not p:
                info = ('Access denied for %s because '
                        'its container, %s, has no security '
                        'assertions.' %
                        (cleanupName(name, value),
                         getCleanPath(container, value)))
                logProtection(info, containerbase is accessedbase)
                if (containerbase is accessedbase):
                    raise 'Unauthorized', info
                return 0

            if roles is _noroles: return 1

            # We are going to need a security-aware object to pass
            # to allowed(). We'll use the container.
            value=container

        # Short-circuit tests if we can:
        try:
            if roles is None or 'Anonymous' in roles: return 1
        except TypeError:
            # 'roles' isn't a sequence
            LOG('Zope Security Policy', PROBLEM, "'%s' passed as roles"
                " during validation of '%s' is not a sequence." % (
                `roles`, name))
            raise

        # Check executable security
        stack=context.stack
        if stack:
            eo=stack[-1]

            # If the executable had an owner, can it execute?
            owner=eo.getOwner()
            if (owner is not None) and not owner.allowed(value, roles):
                # We don't want someone to acquire if they can't
                # get an unacquired!
                if len(roles) < 1:
                    info = ('The object "%s", which is contained in '
                           '%s, is marked as private.' % (
                        cleanupName(name, value),
                        getCleanPath(container, value)))
                elif userHasRolesButNotInContext(owner, value, roles):
                    info = ('User %s, who is the owner of %s, is '
                           'defined in the context of %s, which is not '
                           'an ancestor of %s. '
                            % (
                        owner.getUserName(), getCleanPath(eo),
                        getCleanPath(owner.aq_parent.aq_parent),
                        cleanupName(name, value),
                        #aqChain(owner.aq_parent.aq_parent),
                        #aqChain(value),
                        ))
                else:
                    info = ('User %s, who is the owner of %s, is not '
                           'allowed to access '
                           '%s, which is contained in %s. The owner, '
                           'whose roles are %s, would need to have '
                           'one of the following roles: %s.' % (
                        owner.getUserName(),
                        getCleanPath(eo), cleanupName(name, value),
                        getCleanPath(container, value),
                        tuple(owner.getRolesInContext(value)),
                        tuple(roles)))
                logProtection(info, containerbase is accessedbase)
                if accessedbase is containerbase:
                    raise 'Unauthorized', info
                return 0

            # Proxy roles, which are a lot safer now.
            proxy_roles=getattr(eo, '_proxy_roles', None)
            if proxy_roles:
                for r in proxy_roles:
                    if r in roles: return 1

                # Proxy roles actually limit access!
                if len(roles) < 1:
                    info = ('The object %s, which is contained in '
                           '"%s", is marked as private.' % (
                        cleanupName(name, value),
                        getCleanPath(container, value)))
                else:
                    info = ('The proxy roles set for '
                           '%s do not allow access to %s, which is '
                           'contained in %s. The proxy roles would '
                           'have to include one of the following '
                           'roles: %s' % (
                        getCleanPath(eo), cleanupName(name, value),
                        getCleanPath(container, value),
                        tuple(roles)))
                logProtection(info, containerbase is accessedbase)
                if accessedbase is containerbase:
                    raise 'Unauthorized', info
                return 0


        try:
            if context.user.allowed(value, roles): return 1
        except AttributeError: pass

        # We don't want someone to acquire if they can't get an unacquired!
        if len(roles) < 1:
            info = ('The object %s, '
                   'which is contained in %s, is marked as private.' % (
                cleanupName(name, value),
                getCleanPath(container, value)))
        elif userHasRolesButNotInContext(context.user, value, roles):
            info = ('User %s is '
                    'defined in the context of %s, which is not '
                    'an ancestor of %s.'
                    % (
                context.user.getUserName(),
                getCleanPath(context.user.aq_parent.aq_parent),
                cleanupName(name, value),
                #aqChain(context.user.aq_parent.aq_parent),
                #aqChain(value),
                ))
        else:
            info = ('User %s is '
                    'not allowed to access %s, which is contained '
                    'in %s. The user, whose roles are %s, would '
                    'need to have one of the following roles: %s.' % (
                context.user.getUserName(), cleanupName(name, value),
                getCleanPath(container, value),
                tuple(context.user.getRolesInContext(value)),
                tuple(roles)))
        logProtection(info, containerbase is accessedbase)
        if accessedbase is containerbase:
            raise 'Unauthorized', info
        return 0

    def checkPermission(self, permission, object, context):
        roles=rolesForPermissionOn(permission, object)
        if roles is _what_not_even_god_should_do: return 0
        return context.user.has_role(roles, object)


def cleanupName(name, value):
    # If value is available, does getCleanPath().
    if value is not None:
        return getCleanPath(value)
    return name

def getCleanPath(obj, child=None):
    if obj is None:
        if hasattr(child, 'im_self'):
            obj = getattr(child, 'im_self')
        else:
            child = aq_inner(child)
            parent = aq_parent(child)
            if parent is not None:
                obj = parent
            else:
                obj = child
    name = getPathOf(obj)
    return name

def userHasRolesButNotInContext(user, object, object_roles):
    '''Returns 1 if the user has any of the listed roles but
    is not defined in a context which is not an ancestor of object.
    '''
    if object_roles is None or 'Anonymous' in object_roles:
        return 0
    usr_roles=user.getRolesInContext(object)
    for role in object_roles:
        if role in usr_roles:
            # User has the roles.
            return (not verifyAcquisitionContext(
                user, object, object_roles))
    return 0

def verifyAcquisitionContext(self, object, object_roles=None):
    """Mimics the relevant section of User.allowed(). self is a user object.
    """
    ufolder = aq_parent(self)
    ucontext = aq_parent(ufolder)
    if ucontext is not None:
        if object is None:
            # This is a strange rule, though
            # it doesn't cause any security holes. SDH
            return 1
        if not hasattr(object, 'aq_inContextOf'):
            if hasattr(object, 'im_self'):
                # This is a method.  Grab its self.
                object=object.im_self
            if not hasattr(object, 'aq_inContextOf'):
                # object is not wrapped, therefore we
                # can't determine context.
                # Fail the access attempt.  Otherwise
                # this would be a security hole.
                return None
        if not object.aq_inContextOf(ucontext, 1):
            if 'Shared' in object_roles:
                # Damn, old role setting. Waaa
                object_roles=self._shared_roles(object)
                if 'Anonymous' in object_roles: return 1
            return None
    # Note that if self were not wrapped, it would
    # not be possible to determine the user's context
    # and this method would return 1.
    # However, as long as user folders always return
    # wrapped user objects, this is safe.
    return 1

def aqChain(object):
    # Returns a representation of the acquisition chain of an object.
    obj = object
    rval = []
    while obj is not None:
        base = aq_base(obj)
        rval.append(type(base))
        obj = aq_inner(aq_parent(obj))
    rval.append(object.aq_chain)
    return rval


from zLOG import LOG, INFO

def logProtection(info, exception_raised=0):
    #if exception_raised:
    #    LOG('Security', INFO, info)
    pass

